/*
 *       File:    object.c
 *       Purpose: object extension to Tcl
 *       Author:  Copyright (c) 1998 Peter De Rijk
 *
 *       See the file "README" for information on usage and redistribution
 *       of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "object.h"
#include <string.h>
#include "tcl.h"
#include "class.h"
#include "method.h"

Classy_Method Classy_MethodClassMethod;

void Classy_FreeObject(char *clientdata) {
	Object *object = (Object *)clientdata;
	if (object->name != NULL) {
		Tcl_DecrRefCount(object->name);
		object->name = NULL;
	}
	if (object->trace != NULL) {
		Tcl_DecrRefCount(object->trace);
		object->trace = NULL;
	}
	Tcl_Free(clientdata);
}

int Classy_ObjectObjCmd(
	ClientData clientdata,
	Tcl_Interp *interp,
	int argc,
	Tcl_Obj *argv[])
{
	Class *aclass;
	Object *object;
	Classy_HashEntry *entry;
	Tcl_Obj *cmdObj;
	char *cmd;
	int error;

	object = (Object *)clientdata;
	aclass = object->parent;
	if (argc<2) {
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp,"no value given for parameter \"cmd\" to \"",Tcl_GetStringFromObj(argv[0],NULL),"\"", (char *)NULL);
		return TCL_ERROR;
	}
	if (object->trace != NULL) {
		Tcl_Obj *cmd;
		cmd = Tcl_DuplicateObj(object->trace);
		Tcl_IncrRefCount(cmd);
		error = Tcl_ListObjAppendElement(interp, cmd, Tcl_NewListObj(argc,argv));
		if (error != TCL_OK) {Tcl_DecrRefCount(cmd);return error;}
		error = Tcl_EvalObj(interp,cmd);
		Tcl_DecrRefCount(cmd);
		if (error != TCL_OK) {return error;}
	}
	cmdObj = argv[1];
	cmd = Tcl_GetStringFromObj(cmdObj,NULL);
	argv+=2;
	argc-=2;
	entry = Classy_FindHashEntry(&(aclass->methods), cmdObj);
	if (entry == NULL) {
		error = Tcl_VarEval(interp,"auto_load ::Class::",Tcl_GetStringFromObj(aclass->aclass,NULL),",,m,",cmd,(char *)NULL);
		if (error) {return error;}
		entry = Classy_FindHashEntry(&(aclass->methods), cmdObj);
	}
	if (entry != NULL) {
		Method *method;
		method = (Method *)Classy_GetHashValue(entry);
		Tcl_Preserve((ClientData)object);
		error = Classy_ExecMethod(interp,method,aclass,object,argc,argv);
		if (error != TCL_OK) {
			Tcl_Obj *errorObj = Tcl_NewStringObj("\nwhile invoking method \"",24);
			Tcl_AppendStringsToObj(errorObj, cmd, "\" of object \"", Tcl_GetStringFromObj(object->name,NULL), "\"\n", (char *) NULL);
			Tcl_AddObjErrorInfo(interp, Tcl_GetStringFromObj(errorObj,NULL), -1);
			Tcl_DecrRefCount(errorObj);
		}
		Tcl_Release((ClientData)object);
		return error;
	} else {
		Tcl_Obj *result, **objv;
		int objc,i;
		Tcl_ResetResult(interp);
		error = Classy_InfoMethods(interp,aclass,NULL);
		if (error != TCL_OK) {return error;}
		result = Tcl_GetObjResult(interp);
		Tcl_IncrRefCount(result);
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp,"bad option \"",cmd ,"\": must be ", (char *)NULL);
		error = Tcl_ListObjGetElements(interp, result, &objc, &objv);
		if (error != TCL_OK) {return error;}
		Tcl_AppendElement(interp, Tcl_GetStringFromObj(objv[0],NULL));
		for (i=1;i<objc;i++) {
			Tcl_AppendResult(interp, ", ", Tcl_GetStringFromObj(objv[i],NULL), NULL);
		}
		Tcl_DecrRefCount(result);
		return TCL_ERROR;
	}
}

void Classy_ObjectDestroy(ClientData clientdata) {
	Class *aclass, *tempclass;
	Object *object;
	Classy_HashEntry *entry;
	char *string;
	object = (Object *)clientdata;
	Tcl_Preserve(clientdata);
	aclass = object->parent;
	tempclass = aclass;
	while(1) {
		if (tempclass->destroy != NULL) {
			Classy_ExecMethod(aclass->interp,tempclass->destroy,aclass,object,0,NULL);
		}
		tempclass = tempclass->parent;
		if (tempclass == NULL) break;
	}
	string = Tcl_GetStringFromObj(object->name,NULL);
	entry = Classy_FindHashEntry(&(aclass->children),object->name);
	if (entry != NULL) {Classy_DeleteHashEntry(entry);}
	Tcl_VarEval(aclass->interp,"foreach ::Class::var [info vars ::Class::", string,",,*] {unset $::Class::var}", (char *)NULL);
	Tcl_EventuallyFree(clientdata,Classy_FreeObject);
	Tcl_Release(clientdata);
}

int Classy_ObjectDestroyObjCmd(
	Tcl_Interp *interp,
	Class *aclass,
	Object *object,
	int argc,
	Tcl_Obj *CONST argv[])
{
	Tcl_Preserve((ClientData)object);
	Tcl_DeleteCommandFromToken(interp,object->token);
	Tcl_Release((ClientData)object);
	return TCL_OK;
}

static Classy_HashTable classcurrent;
void Classy_InitSuper() {
	Classy_InitHashTable(&classcurrent);
}

int Classy_NewClassMethod(
	Tcl_Interp *interp,
	Class *aclass,
	Object *object,
	int argc,
	Tcl_Obj *CONST argv[])
{
	Class *tempclass;
	Classy_HashEntry *entry;
	Tcl_Obj *name,*tempobj;
	Tcl_CmdInfo cmdinfo;
	char *string;
	int len,pos,found,error,anew;

	if (argc<1) {
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp,"wrong # args: should be \"",Tcl_GetStringFromObj(aclass->aclass,NULL),
			" new object ?args?\"", (char *)NULL);
		return TCL_ERROR;
	}
	string = Tcl_GetStringFromObj(argv[0],&len);
	if (len>2) {
		if ((string[0]==':')&&(string[1]==':')) {
			string += 2;
			len -= 2;
		}
		pos = len - 1;
		while (pos>1) {
			if ((string[pos]==':')&&(string[pos-1]==':')) break;
			pos--;
		}
		if (pos != 1) {
			string[pos-1] = '\0';
			error = Tcl_VarEval(interp,"namespace eval ",string, " {}", (char *)NULL);
			if (error != TCL_OK) {return error;}
			error = Tcl_VarEval(interp,"namespace eval ::Class::",string, " {}", (char *)NULL);
			if (error != TCL_OK) {return error;}
			string[pos-1] = ':';
		}
	}
	name = Tcl_NewStringObj(string,len);
	found = Tcl_GetCommandInfo(interp,string,&cmdinfo);
	if (found == 1) {
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp,"command \"",string,"\" exists", (char *)NULL);
		return TCL_ERROR;
	}
	entry = Tcl_CreateHashEntry(&(aclass->children),name,&anew);
	if (anew == 1) {
		object = (Object *)Tcl_Alloc(sizeof(Object));
		Classy_SetHashValue(entry,(ClientData)object);
	} else {
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp,"command \"",string,"\" exists", (char *)NULL);
		return TCL_ERROR;
	}
	object->parent = aclass;
	object->name = name;
	Tcl_IncrRefCount(name);
	object->trace = NULL;
	object->token = Tcl_CreateObjCommand(interp,string,(Tcl_ObjCmdProc *)Classy_ObjectObjCmd,
		(ClientData)object,(Tcl_CmdDeleteProc *)Classy_ObjectDestroy);
	tempclass = aclass;
	while(1) {
		if (tempclass->init != NULL) {
			Tcl_Obj *stringkey;
			Tcl_ResetResult(interp);
			stringkey = Tcl_NewStringObj(string,len);
			Tcl_AppendToObj(stringkey,",init",5);
			entry = Tcl_CreateHashEntry(&classcurrent,stringkey,&anew);
			Classy_SetHashValue(entry, (ClientData)tempclass);
			error = Classy_ExecMethod(interp,tempclass->init,aclass,object,argc-1,argv+1);
			Classy_DeleteHashEntry(entry);
			if (error != TCL_OK) {
				Tcl_Obj *errorObj, *errorinfo;
				errorObj = Tcl_GetObjResult(interp);
				Tcl_IncrRefCount(errorObj);
				tempobj = Tcl_NewStringObj("errorInfo",-1);
				errorinfo = Tcl_ObjGetVar2(interp, tempobj, NULL, TCL_GLOBAL_ONLY);
				Tcl_DecrRefCount(tempobj);
				Tcl_IncrRefCount(errorinfo);
				Tcl_DeleteCommandFromToken(interp,object->token);
				Tcl_ResetResult(interp);
				Tcl_AddObjErrorInfo(interp, Tcl_GetStringFromObj(errorinfo,NULL), -1);
				Tcl_SetObjResult(interp,errorObj);
				Tcl_DecrRefCount(errorObj);
				Tcl_DecrRefCount(errorinfo);
				return error;
			}
			break;
		}
		tempclass = tempclass->parent;
		if (tempclass == NULL) {
		Tcl_SetObjResult(interp,name);
			break;
		}
	}
	return TCL_OK;
}

int Classy_TraceMethod(
	Tcl_Interp *interp,
	Class *aclass,
	Object *object,
	int argc,
	Tcl_Obj *CONST argv[])
{
	char *command;
	int len;

	if (argc!=1) {
		Tcl_ResetResult(interp);
		Tcl_WrongNumArgs(interp,argc,argv," procedure");
		return TCL_ERROR;
	}
	command = Tcl_GetStringFromObj(argv[0],&len);
	if (len != 0) {
		if (object != NULL) {
			object->trace = Tcl_DuplicateObj(argv[0]);
			Tcl_IncrRefCount(object->trace);
		} else {
			aclass->trace = Tcl_DuplicateObj(argv[0]);
			Tcl_IncrRefCount(aclass->trace);
		}
	} else {
		if (object != NULL) {
			if (object->trace != NULL) {
				Tcl_DecrRefCount(object->trace);
				object->trace = NULL;
			}
		} else {
			if (aclass->trace != NULL) {
				Tcl_DecrRefCount(aclass->trace);
				aclass->trace = NULL;
			}
		}
	}
	return TCL_OK;
}

int Classy_SuperObjCmd(
	ClientData clientdata,
	Tcl_Interp *interp,
	int argc,
	Tcl_Obj *CONST argv[])
{
	Tcl_CmdInfo cmdinfo;
	Classy_HashEntry *entry, *mentry;
	Class *aclass, *current;
	Object *object = NULL;
	Method *cmethod, *bmethod;
	CONST char *classname;
        CONST char *objectname;
        char *option;
	Tcl_Obj *stringkey, *optionObj;
	int found, error, optlen, anew;

	if (argc<1) {
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp,"wrong # args: should be \"super option ...\"", (char *)NULL);
		return TCL_ERROR;
	}
	classname = Tcl_GetVar(interp, "class", TCL_LEAVE_ERR_MSG);
	if (classname == NULL) {
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp,"no variable \"class\" found", (char *)NULL);
		return TCL_ERROR;
	}
	found = Tcl_GetCommandInfo(interp, classname, &cmdinfo);
	if (found == 0) {
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp,"class \"",classname,"\" does not exist", (char *)NULL);
		return TCL_ERROR;
	}
	aclass = (Class *)cmdinfo.objClientData;
	objectname = Tcl_GetVar(interp, "object", TCL_LEAVE_ERR_MSG);
	optionObj = argv[1];
	option = Tcl_GetStringFromObj(optionObj,&optlen);
	if (objectname == NULL) {
		if ((optlen == 4) &&(strncmp(option,"init",4)==0)) {
			Tcl_ResetResult(interp);
			Tcl_AppendResult(interp,"no variable \"object\" found", (char *)NULL);
			return TCL_ERROR;
		} else {
			stringkey = Tcl_NewStringObj(classname,strlen(classname));
			Tcl_AppendToObj(stringkey,",",1);
			Tcl_AppendToObj(stringkey,option,optlen);
		}
	} else {
		found = Tcl_GetCommandInfo(interp, objectname, &cmdinfo);
		if (found == 0) {
			Tcl_ResetResult(interp);
			Tcl_AppendResult(interp,"object \"",objectname,"\" does not exist", (char *)NULL);
			return TCL_ERROR;
		}
		object = (Object *)cmdinfo.objClientData;
		stringkey = Tcl_NewStringObj(objectname,strlen(objectname));
		Tcl_AppendToObj(stringkey,",",1);
		Tcl_AppendToObj(stringkey,option,optlen);
	}
	entry = Tcl_CreateHashEntry(&classcurrent,stringkey,&anew);
	if (anew == 0) {
		current = (Class *)Classy_GetHashValue(entry);
	} else {
		current = aclass;
	}

	if ((optlen == 4) &&(strncmp(option,"init",4)==0)) {
		if (current->parent != NULL) {
			current = current->parent;
			while(1) {
				if (current->init != NULL) {
					Classy_SetHashValue(entry, (ClientData)current);
					error = Classy_ExecMethod(interp,current->init,aclass,object,argc-2,argv+2);
					if (anew != 0) {Classy_DeleteHashEntry(entry);}
					return error;
				}
				current = current->parent;
				if (current == NULL) break;
			}
		}
		Tcl_SetResult(interp,(char*)objectname,TCL_VOLATILE);
	} else if (objectname != NULL) {
		if (current->parent == NULL) {
			Tcl_ResetResult(interp);
			Tcl_AppendResult(interp,"No method \"",option,"\" defined for super of ",objectname,
				" (at class \"",Tcl_GetStringFromObj(current->aclass,NULL),"\"", (char *)NULL);
			if (anew != 0) {Classy_DeleteHashEntry(entry);}
			return TCL_ERROR;
		}
		mentry = Classy_FindHashEntry(&(current->methods), optionObj);
		if (mentry != NULL) {
			cmethod = (Method *)Classy_GetHashValue(mentry);
			current = current->parent;
			while(1) {
				mentry = Classy_FindHashEntry(&(current->methods), optionObj);
				if (mentry == NULL) break;
				bmethod = (Method *)Classy_GetHashValue(mentry);
				if (cmethod != bmethod) {
					Classy_SetHashValue(entry, (ClientData)current);
					Tcl_Preserve((ClientData)object);
					Classy_SetHashValue(entry, (ClientData)current);
					error = Classy_ExecMethod(interp,bmethod,aclass,object,argc-2,argv+2);
					Tcl_Release((ClientData)object);
					if (anew != 0) {Classy_DeleteHashEntry(entry);}
					return error;
				}
				if (current->parent == NULL) break;
				current = current->parent;
			}
		}
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp,"No method \"",option,"\" defined for super of ",objectname,
			" (at class \"",Tcl_GetStringFromObj(current->aclass,NULL),"\")", (char *)NULL);
		if (anew != 0) {Classy_DeleteHashEntry(entry);}
		return TCL_ERROR;
	} else {
		if (current->parent == NULL) {
			Tcl_ResetResult(interp);
			Tcl_AppendResult(interp,"No classmethod \"",option,"\" defined for super of ",classname,
				" (at class \"",Tcl_GetStringFromObj(current->aclass,NULL),"\"", (char *)NULL);
			if (anew != 0) {Classy_DeleteHashEntry(entry);}
			return TCL_ERROR;
		}
		mentry = Classy_FindHashEntry(&(current->classmethods), optionObj);
		if (mentry != NULL) {
			cmethod = (Method *)Classy_GetHashValue(mentry);
			current = current->parent;
			while(1) {
				mentry = Classy_FindHashEntry(&(current->classmethods), optionObj);
				if (mentry == NULL) break;
				bmethod = (Method *)Classy_GetHashValue(mentry);
				if (cmethod != bmethod) {
					Classy_SetHashValue(entry, (ClientData)current);
					Tcl_Preserve((ClientData)aclass);
					Classy_SetHashValue(entry, (ClientData)current);
					error = Classy_ExecClassMethod(interp,bmethod,aclass,NULL,argc-2,argv+2);
					Tcl_Release((ClientData)aclass);
					if (anew != 0) {Classy_DeleteHashEntry(entry);}
					return error;
				}
				if (current->parent == NULL) break;
				current = current->parent;
			}
		}
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp,"No classmethod \"",option,"\" defined for super of ",classname,
			" (at class \"",Tcl_GetStringFromObj(current->aclass,NULL),"\")", (char *)NULL);
		if (anew != 0) {Classy_DeleteHashEntry(entry);}
		return TCL_ERROR;
	}
	
	if (anew != 0) {Classy_DeleteHashEntry(entry);}
	return TCL_OK;
}


