
# Modul: c.tcl
#
# Vervollstndigt die Schnittstelle zu der generierten c-Datenstruktur
# um weitere Funktionen.


# Hier die Basis-Funktionen, die in C implementiert sind.
# Siehe auch:
#  modelGen.tclinter
#  ObservedData.tclinter
#
# Fr alle Sprachelemente:
#   XXX_new
#   XXX_type
#
# Von ObservedData.fw:
#
# VLString getNameOfType(LanguageElementTypeDescriptor *x) -> METHOD;
# VLString getSubtypes(LanguageElementTypeDescriptor *x) -> METHOD;
# VLString getSupertypes(LanguageElementTypeDescriptor *x) -> METHOD;
# VLInt getIsAbstract(LanguageElementTypeDescriptor *x) -> METHOD;
# VLBoolean isSubtypeOf(LanguageElementTypeDescriptor *x, LanguageElementTypeDescriptor *other) -> METHOD;
#
# ChangeHandler *ChangeHandler_new(VLValue cmd) -> CONSTRUCTOR ChangeHandler;
#
# VLBoolean isLanguageObject(StructureObject *obj) -> METHOD;
# VLBoolean isVirtualLanguageObject(StructureObject *obj) -> METHOD;
# VLBoolean isEOLObject(StructureObject *obj) -> METHOD;
# VLBoolean isRelationObject(StructureObject *obj) -> METHOD;
#
# VLString getClassName(StructureObject *obj) -> METHOD;
# VLList getSubElements(StructureObject *obj) -> METHOD;
# StructureObject *getParent(StructureObject *obj) -> METHOD;
# void addChangeHandler(StructureObject *obj, ChangeHandler *x) -> METHOD;
# void removeChangeHandler(StructureObject *obj, ChangeHandler *x) -> METHOD;
# void propagateChange(StructureObject *obj) -> METHOD; 
# VLString getInfo(StructureObject *obj) -> METHOD;
# StructureObject *getRoot(StructureObject *obj) -> METHOD;
# VLBoolean isAncestorOf(StructureObject *a, StructureObject *b) -> METHOD;
#
# void insertBefore_internal(VirtualLanguageObject *obj, LanguageObject *elem) -> METHOD insertBefore;
# VirtualLanguageObject *getSuccessor(VirtualLanguageObject *obj) -> METHOD;
# VirtualLanguageObject *getPredecessor(VirtualLanguageObject *obj) -> METHOD;
#
# VLBoolean canExtract(LanguageObject *obj) -> METHOD;
# void extract(LanguageObject *obj) -> METHOD;
# LanguageElementTypeDescriptor *getTypeId(LanguageObject *obj) -> METHOD;
# VLList getRefererList(LanguageObject *obj) -> METHOD;
#
# VLString getRole(RelationObject *obj) -> METHOD;
# VLBoolean isAggregate(RelationObject *obj) -> METHOD;
#
# LanguageElementTypeDescriptor *getAcceptedType(AggregateRelation *obj) -> METHOD;
# EOLObject *getEOL(AggregateRelation *obj) -> METHOD;
# VirtualLanguageObject *getFirst(AggregateRelation *obj) -> METHOD;
# VLBoolean maxMultiplicityIsOne(AggregateRelation *obj) -> METHOD;
# VLBoolean minMultiplicityIsOne(AggregateRelation *obj) -> METHOD;
# VLBoolean minProcessorMultiplicityIsOne(AggregateRelation *obj) -> METHOD;
#
# VLValue getValue(ObservedType *obj) -> METHOD;
# void setValue_internal(ObservedType *obj, VLValue value) -> METHOD setValue;
# VLString getTypeName(ObservedType *obj) -> METHOD;

# --- class hierarchy  ----------------------------------------------------------------

# Klassenzugehrigkeit kann mit den Abfragefunktionen isXXX abgefragt werden.
#
# -StructureObject
#   -VirtualLanguageObject
#     -LanguageObject ("objectNode")
#     -EOLObject
#   -RelationObject ("attrNode")
#     -Leaf
#       -Value
#       -Reference
#     -Aggregate



proc c::isStructureObject {obj} {
    # for the sake of completeness
    return 1
}


proc c::isLeaf {obj} {
    return [expr { [c::getClassName $obj] eq "ObservedType" }]
}


proc c::isTerminal {obj} {
    # OBSOLETE!
    return [isLeaf $obj]
}


proc c::isValue {obj} {
    return [expr {[c::isLeaf $obj] && [c::getTypeName $obj] ne "VLPointer" }]
}


proc c::isReference {obj} {
    return [expr {[c::isTerminal $obj] && [c::getTypeName $obj] eq "VLPointer" }]
}


proc c::isAggregate {obj} {
    return [expr { [c::getClassName $obj] eq "AggregateRelation" }]
}



# --- insertion contexts  ------------------------------------------------------------


proc c::icBelow {aggObj} {
    return [getEOL $aggObj]
}

proc c::icAfter {elem} {
    return [getSuccessor $elem]
}

proc c::icBefore {elem} {
    return $elem
}


# --- pretty printing ----------------------------------------------------------------

proc c::getContext {obj} {
    if { [c::isLanguageObject $obj] } {
        return [c::getClassName $obj]
    } elseif { [c::isRelationObject $obj] } {
        return "[c::getClassName [c::getParent $obj]].[c::getRole $obj]"
    } elseif { [c::isEOLObject $obj] } {
        return "EOL_OBJECT"
    } else {
        return "UNKNOWN_OBJECT"
    }
}


# Better name: getSuperclasses
proc c::getContexts {node} {
    set result ""
    if { [c::isLanguageObject $node] } {
        set obj $node
        set role ""
    } else {
        set obj [getParent $node]
        set role ".[c::getRole $node]"
    }
    foreach type [getSupertypes [getTypeId $obj]] {
        lappend result "[getNameOfType $type]$role"
    }
    return $result
}


proc c::getObjectRep {node} {
    if { [isLanguageObject $node] } {
        return "[getContext $node] '[basiccomp::getFullName $node]'"
    } elseif { [c::isTerminal $node] } {
        set class [c::getClassName [c::getParent $node]]
        set name [basiccomp::getFullName [c::getParent $node]]
        return "ATTR [c::getRole $node] OF $class '$name'"
    } else {
        return "XXX"
    }
}


proc c::getComment {obj} {
    if { [isVirtualLanguageObject $obj] } {
        if { [getSubElement $obj name -var nameObj] } {
            set name [getValue $nameObj]
            if { $name eq "" } {
                set name "~"
            }
        } else {
            set name "\#"
        }
        return "${name}([getClassName $obj])"
    } elseif { [isRelationObject $obj] } {
        return "[getComment [getParent $obj]].[getRole $obj]([getClassName $obj])"
    } else {
        xn::assert { 0 }
    }
}


proc c::getCommentedObject_unused {obj} {
    return "$obj {[getComment $obj]}"
}


proc c::print_unused {elem {prefix ""}} {
    if { [c::isLanguageObject $elem] } {
        puts "$prefix [c::getClassName $elem]"
    } else {
        puts "$prefix [c::getRole $elem] ([c::getInfo $elem])"
    }
    foreach sub [c::getSubElements $elem] {
        print $sub "$prefix  "
    }
}



# ---- Persistent IDs -----------------------------------------------------------------------------------


# PUBLIC
proc c::setPersistentId {file node id} {
    variable data
    if { ![info exists data(node2id,$file-$node)] && ![info exists data(id2node,$file-$id)] } {
        set data(node2id,$file-$node) $id
        set data(id2node,$file-$id) $node
    }
}


# PUBLIC
proc c::getPersistentId {file node} {
    variable data
    if { ![info exists data(node2id,$file-$node)] } {
        set id [getNewPersistentId $file]
        set data(node2id,$file-$node) $id
        set data(id2node,$file-$id) $node
    }
    return $data(node2id,$file-$node)
}


# PRIVATE
proc c::getNewPersistentId {file} {
    variable data
    set maxId 1000000
    while 1 {
        set candidate [expr { $maxId + int($maxId * rand()) }]
        if { ![info exists data(id2node,$file-$candidate)] } {
            return $candidate
        }
    }
}


# --- get-Functions ------------------------------------------------------------------

# Severity:
# ---------
# 1. Null-Ref, Fehlendes Including
# 2: Nicht existierendes Attribut, Child liefert mehr als einen Wert
# 3: Syntaxfehler


proc c::catchErrLevel {errLevel cmd exceptionHandler} {
  # fhrt $cmd aus
  # Falls ein Fehler <= $errLevel auftritt wird $expeptionHandler ausgefhrt
  
    set err [catch {uplevel 1 $cmd} errMsg]
    switch $err {
        0 {
            return $errMsg
        }
        1 {
            set savedInfo $::errorInfo
            if { ![regexp {^Severity ([0-9]+):} $errMsg dummy severity] } {
                set severity 99	  
            }
            if { $severity > $errLevel } {
                error $errMsg $savedInfo
            } else {
                uplevel 1 $exceptionHandler
            }
        }
        2 {
            return -code return $errMsg 
        }
        default {
            error "xxx"
        }
    }
}


proc c::changeException {newMessage cmd} {
    set err [catch {uplevel 1 $cmd} errMsg]
    switch $err {
        0 {
            return $errMsg
        }
        1 {
            set savedInfo $::errorInfo
            error "$newMessage: $errMsg" $savedInfo
            #return -code error "$newMessage: $errMsg"
        }
        2 {
            return -code return $errMsg 
        }
        default {
            error "xxx"
        }
    }
}


proc c::_getList {path} {
    # Liefert das Ergebnis des Pfadausdrucks $path (ohne Substitution)
    # Bei einem Fehler wird eine Exception geworfen

    #regexp {([^.]+).(.+)} $path dummy obj tail
    regexp {^(0x[0-9a-fA-Z]+)(.*)} $path dummy obj tail
    return [treePath::eval $obj "THIS $tail"]
}


proc c::_get {path} {
    # Liefert das Ergebnis des Pfadausdrucks $path (ohne Substitution)
    # Bei einem Fehler wird eine Exception geworfen
    
    return [evalSimplePathExpression2 $path]
}


proc c::_get_thisNotation {obj path} {
    # Transforms into other format
    regexp {THIS\.(.+)} $path dummy tail
    return [_get $obj.$tail]
}


proc c::get {path {errLevel 0} {defaultValue ""}} {
    # Falls der Pfadausdrucks $path (mit Substitution) evaluiert werden kann
    # wird das Ergebnis des Pfadausdrucks zurckgeliefert.
    # Falls beim Auswerten von $path ein Fehler <= $errLevel auftritt wird
    # $defaultValue zurckgeliefert.
    # Falls beim Auswerten von $path ein Fehler > $errLevel auftritt wird 
    # eine Exception ausgelst.

    set path [uplevel 1 [list subst -nobackslashes -nocommands $path]]
    changeException "Error while evaluating $path" {
        catchErrLevel $errLevel {
            return [_get $path]
        } {
            return $defaultValue
        }
    }
}


proc c::get_thisNotation {obj path {errLevel 0} {defaultValue ""}} {
    changeException "Error while evaluating $path" {
        catchErrLevel $errLevel {
            return [_get_thisNotation $obj $path]
        } {
            return $defaultValue
        }
    }
}


proc c::tryGet {path resultVar {errLevel 1}} {
    # Falls der Pfadausdrucks $path (mit Substitution) evaluiert werden kann
    # wird 1 zurckgegeben und das Ergebnis des Pfadausdrucks in resultVar gespeichert.
    # Falls beim Auswerten von $path ein Fehler <= $errLevel auftritt wird 0 zurckgegeben.
    # Falls beim Auswerten von $path ein Fehler > $errLevel auftritt wird eine Exception ausgelst.

    set path [uplevel 1 [list subst -nobackslashes -nocommands $path]]
    upvar $resultVar result

    changeException "Error while evaluating $path" {
        catchErrLevel $errLevel {
            set result [_get $path]
            return 1
        } {
            return 0
        }
    }
}

proc c::getOrNull {path {errLevel 1}} {
    # obsolete!
    return [uplevel 1 [list c::get $path $errLevel 0x0]]
}    


proc c::getList {path} {
    # Liefert das Ergebnis des Pfadausdrucks $path (mit Substitution)

    set path [uplevel 1 [list subst -nobackslashes -nocommands $path]]

    changeException "Error while evaluating $path" {
        return [_getList $path]
    }
}


# --- Rest ---------------------------------------------------------------------------


proc c::startProcessor {name root} {
    #if {[info exists ::c::processor_running]} {
    #    error "LIGA Processor is already running!"
    #}
    #set ::c::processor_running 1
    vlProperty::resetForEvaluator
    vlAccess::setRootNode $root
    startProcessor_$name $root
    c::cleanupTemp
    #unset ::c::processor_running
}


proc c::findReferences {targetObj root} {
    set result ""
    foreach ref [getRefererList $targetObj] {
        if { [isAncestorOf $root $ref] } {
            lappend result $ref
        }
    }
    return $result
}


proc c::getAttr {obj name} {
    return [getValue [getSubElement $obj $name]]
}


proc c::getAggregateSubElements {obj} {
    set result ""
    foreach sub [c::getSubElements $obj] {
        if { [c::isAggregate $sub] } {
            lappend result $sub
        }
    }
    return $result
}


xn::proc c::getSubElement {obj name -var ""} {
    log::log XXX "in C::getSubElement $name"
    if { [nullptr $obj] } {
        error "Severity 1: NULL Pointer while accessing attribute '$name'"
    }

    if { $var ne "" } {
        upvar $var result
    }
    set result ""
    log::log XXX "in C::getSubElement $name vor geSubElements"
    set subs [c::getSubElements $obj]
    log::log XXX "in C::getSubElement $name nach geSubElements"
    foreach sub $subs {
        log::log XXX "in FOREACH SUB: $sub"
        if { [string compare $name [c::getRole $sub]]==0 } {
            log::log XXX "in IF"
            set result $sub
            break
        }
        log::log XXX "nach IF"
   }
   log::log XXX "in C::getSubElement $name vor IF"
    if { $var eq "" } {
        if { $result eq "" } {
           error "Severity 2: No sub element '$name' in '[c::getClassName $obj]'"
        } else {
            return $result
        }
    } else {
        return [expr { $result ne "" }]
    }
}


proc c::nodeIsSubtypeOf {node type} {
    if { ![isLanguageObject $node] } {
        return 0
    }
    return [isSubtypeOf [getTypeId $node] [${type}_type]]
}


proc c::findIncluding {type node} {
    set orig $node
    if { ![nullptr $node] } {
        set node [c::getParent $node]
    }        
    while 1 {
        if { [nullptr $node] } {
            error "Severity 1: Can't find including $type of node $orig"
        }
        if { [nodeIsSubtypeOf $node $type] } {
            return $node
        }
        set node [c::getParent $node]
    }
}


proc c::evalSimplePathExpression {obj path} {
    # Format: THIS-Style
    # Called by vlGet in vlAccess.c

    # Transforms into other format
    regexp {THIS\.(.+)} $path dummy tail

    set error [catch {
        set result [c::get $obj.$tail]
    } errMessage]
    if { $error } {
        log::error "Error while evaluating '$obj.$tail'. Look for String '***ERROR***' in generated output!"
        set result "***ERROR***"        
    }
    return $result
}   


proc c::evalSimplePathExpression2 {path} {

    set list [split $path "."]
    set curr [lindex $list 0]
    foreach jump [lrange $list 1 end] {
        switch -glob $jump {
            "PARENT" {
                set curr [c::getParent $curr]
            }
            "VALUE" {
                if { [nullptr $curr] } {
                    error "Severity 1: .VALUE applied to NULL-Reference"
                } else {
                    set curr [c::getValue $curr]
                }
            }
            "CHILD" {
                if { [nullptr $curr] } {
                    error "Severity 1: .CHILD applied to NULL-Reference"
                }
                set subList [c::getSubElements $curr]
                if { [llength $subList] > 1 } {
                    error "Severity 2: .CHILD yields more than one value"
                }
                if { [llength $subList] == 1 } {
                    set curr [lindex $subList 0]
                } else {
                    set curr "0x0"
                }
            }
            "BASE" {
                if { [getModel $curr model] } {
                    set curr $model
                } else {
                    error "Severity 1: .BASE yields NULL"
                }                
            }
            "INCLUDING *" {
                set curr [findIncluding [lindex $jump 1] $curr]
            }
            "BASEREF" {
                if { [getSubElement $curr baseRef -var jump] } {
                    set curr $jump
                } else {
                    error "Severity 2: .BASEREF yields NULL"
                }
            }
            "%*" {
                if { [nullptr $curr] } {
                    error "Severity 1: .%attr applied to NULL-Reference"
                }
                set name [string range $jump 1 end]
                set curr [attrComp::computeAttr $curr $name]
            }
            default {
                if { [regexp {^\^(.*)$} $jump dummy type] } {
                    set curr [c::findIncluding [c::getParent $curr]]
                } else {
                    set curr [c::getSubElement $curr $jump]
                }
            }
        }
    }
    return $curr
}



proc c::traverseTreeRec {root classCheck varName op upLevel} {
    if { [$classCheck $root] } {
        upvar $upLevel $varName var
        set var $root
        uplevel $upLevel $op
        if { $var eq "" } {
            set root ""
        }
    }
        if { $root ne "" } {
        foreach sub [c::getSubElements $root] {
            traverseTreeRec $sub $classCheck $varName $op [expr $upLevel+1]
        }
    }
}


proc c::traverseTree {root classCheck varName op} {
    traverseTreeRec $root $classCheck $varName $op 2
}



proc c::getModel {obj resultName} {
    # true  - wenn $obj ein Modellelement repraesentiert. In diesem Fall enthaelt
    #         die Variable $resultName das Modellelement
    # false - sonst

    upvar $resultName result

    if { [nullptr $obj] } {
        return 0
    }
    if { [c::isRelationObject $obj] } {
        if { ![getModel [c::getParent $obj] parmodel] } {
            return 0
        }
        if { ![c::getSubElement $parmodel [c::getRole $obj] -var model] } {
            return 0
        }
        set result $model
        return 1
    } elseif { [c::isLanguageObject $obj] } {
        if { ![c::getSubElement $obj baseRef -var baseRef] } {
            return 0
        }
        set model [c::getValue $baseRef]
        if { [nullptr $model] } {
            return 0
        }
        set result $model
        return 1
    } elseif { [c::isEOLObject $obj] } {
        if { ![getModel [getParent $obj] parentModel] } {
            return 0
        }
        set result [getEOL $parentModel]
        return 1
    } else {
        error "Unknown object type: [c::getClassName $obj]"
    }
}


proc c::getModelClass {repClass} {
    return [extractSubstring {[^_]+_(.*)} $repClass]
}


proc c::getRepClass {modelClass viewName} {
    return "${viewName}_${modelClass}"
}


proc c::siblings {a b} {
    return [expr { [getParent $a] eq [getParent $b] }]
}


proc c::isRoot {obj} {
    return [nullptr [c::getParent $obj]]
}



proc c::getSubElementCount {obj} {
    return [llength [getSubElements $obj]]
}


proc c::getTypeIdFromName {typeName} {
    return [${typeName}_type]
}


proc c::hasClass {class} {
    return [procExists "c::${class}_new"]
}


proc c::canInsertType {insertionContext typeId} {
    set agg [c::getParent $insertionContext]
    return [expr {
      [isSubtypeOf $typeId [getAcceptedType [getParent $insertionContext]]] &&
      !([maxMultiplicityIsOne $agg] && [getSubElementCount $agg]==1)
  }]     
}

proc c::canInsert {insertionContext newElem} {
    return [expr {
      ![isAncestorOf $newElem $insertionContext] &&
      [canInsertType $insertionContext [getTypeId $newElem]]
  }]
}


proc c::getObjCount {objNode} {
    set objCount 0
    traverseTree $objNode isLanguageObject subObjNode {
        incr objCount
    }
    return $objCount    
}


proc c::getClassOfChild {aggNode} {
    # Liefert den Klassennamen eines SUB?-Attributs oder den leeren String
    set sub [c::getList $aggNode.CHILDREN]
    switch [llength $sub] {
        0 {
            return ""
        }
        1 {
            return [getClassName [lindex $sub 0]]
        }
        default {
            error "c::getClassOfChild called for SUB*-Attribute"
        }
    }
}


proc c::getAcceptedClassesForType {typeDescriptor} {
    set result ""
    foreach subtype [getSubtypes $typeDescriptor] {
        if { ![getIsAbstract $subtype] } {
            eval lappend result [getNameOfType $subtype]
        }
    }
    return $result
}


proc c::getAcceptedClassesForChild {aggNode} {
    return [getAcceptedClassesForType [getAcceptedType $aggNode]]
}


proc c::getNonEmptyAggSubElements {obj} {
    set list [getSubElements $obj]
    set result ""
    foreach x $list {
        if { [isAggregate $x] && [getFirst $x]!=[getEOL $x] } {
            lappend result $x
        }
    }
    return $result
}
