
# Modul: edit.tcl
#
# Funktionen zum ndern der abstrakten Struktur.
#
# Das Modul bernimmt folgende Aufgaben:
#
# * Die nderungsfunktionen ndern nicht nur die
#   Sicht-Struktur, sondern ggf. auch die
#   Semantische Struktur.
#
# * Falls fr bestimmte Klassen Editierhooks definiert
#   sind, werden diese aufgerufen.
#
#

namespace eval ::edit {}
namespace eval ::edithook {}

# Hooks:
#
# * create_CLASS
# * setAttr_CLASS
#
# * initValue

# --- change0 -------------------------------------------------------------------------

namespace eval ::change0 {}

proc change0::createRoot {} {
    set obj [change::createRoot]
    doNonContextSensitiveInit $obj
    lappend ::edit::newObjects $obj
    return $obj
}


proc change0::createObject {class file} {
    set obj [change::createObject $class $file]
    doNonContextSensitiveInit $obj
    lappend ::edit::newObjects $obj
    return $obj
}


proc change0::insert {refObj newElem} {
    change::insert $refObj $newElem
}


proc change0::setValue {obj value} {
    change::setValue $obj $value
}


proc change0::delete {obj} {
    change::delete $obj
}


proc change0::extract {obj} {
    change::extract $obj
}


proc change0::doNonContextSensitiveInit {obj} {
    foreach attr [c::getSubElements $obj] {
        tryInitValAttr $attr
        tryInitRefAttr $attr
        tryInitExclamationmarkAttr $attr
    }
}


proc change0::tryInitValAttr {attr} {
    if { ![c::isValue $attr] } {
        return
    }
    if { ![equal [c::getRole $attr] "name"] } {
        foreach context [c::getContexts $attr] {
            if { [info exists edithook::initValue($context)] } {
                setValue $attr $edithook::initValue($context)
                return
            }
        }
    }
}


proc change0::tryInitRefAttr {attr} {
    if { ![c::isReference $attr] } {
        return
    }
    foreach context [c::getContexts $attr] {
        if { [info exists edithook::initValue($context)] } {
            set fullName $edithook::initValue($context)
            set file [d::getFile $attr]
            if { [basiccomp::findByFullName $file $fullName -var refObj] } {
                setValue $attr $refObj
            } else {
                log::error "Can't find object '$fullName'"
            }
            return
        }
    }
}


proc change0::tryInitExclamationmarkAttr {attr} {
    if { ![c::isAggregate $attr] || ![c::minMultiplicityIsOne $attr] } {
        return
    }
    set class [c::getNameOfType [c::getAcceptedType $attr]]
    set obj [change0::createObject $class [d::getFile $attr]]
    edit::insert [c::icBelow $attr] $obj
}



# --- internal functions --------------------------------------------------------------


proc edit::ask {question} {
    return [CDlg::Question $question]
}


set edit::newObjects ""


proc edit::doContextSensitiveInit {} {

    set todo $::edit::newObjects
    set ::edit::newObjects ""
    # exernal hooks could create other new objects!

    foreach obj $todo {
        assignInitialName $obj
        executeCreateHooks $obj
    }

}


proc edit::assignInitialName {obj} {
    if [c::getSubElement $obj name -var nameAttr] {
        if [empty [c::getValue $nameAttr]] {
            change0::setValue $nameAttr [basiccomp::computeInitialName $obj]
        }
    }
}


proc edit::executeCreateHooks {obj} {
    set class [c::getClassName $obj]
    if [procExists "edithook::create_$class"] {
        edithook::create_$class $obj
    }    
}


# --- clipboard -----------------------------------------------------------------------


proc edit::cutIntoClipboard {objNode} {
    clipboard::cut $objNode
}

proc edit::mark {objNode} {
    clipboard::mark $objNode
}

proc edit::copyIntoClipboard {objNode} {
    clipboard::copy $objNode
}



# --- edit functions -------------------------------------------------------------------



proc edit::createRoot {} {
    return [change0::createRoot]
}


proc edit::createObject {class file} {
    return [change0::createObject $class $file]
}


proc edit::createModelElement {rep modelElementVar} {
    # erzeugt ein Modellelement fuer Repraensentant $rep.
    # Liefert true, wenn das mglich ist.
    upvar $modelElementVar modelElement

    set file [d::getFile $rep]
    set modelClass [c::getModelClass [c::getClassName $rep]]
    if [procExists c::${modelClass}_new] {
        set modelElement [createObject $modelClass $file]
        setModel $rep $modelElement
        return 1
    } else {
        return 0
    }
}


proc edit::insert {refObj newElem} {

    if [equal $refObj $newElem] {
        # just ignore this case.
        return
    }

    change0::insert $refObj $newElem

    if [c::getModel [c::getParent $refObj] aggModel] {

        # -> Repraesentanten-Liste
        # Auch das Modell muss noch geaendert werden!
        # Defaultverhalten: Aendere auch das Modell

        # finde entsprechendes Modellelement oder erzeuge ein neues
        if [c::getModel $newElem modelElem] {
            # es existiert schon ein Modellelement;
        } else {
            if ![createModelElement $newElem modelElem] {
                # dies ist ein nur-sicht element.
                # nichts tun!
                return
            }
        }

        # suche Stelle, Einfuegestelle im Modell -> refObj
        while ![equal [c::get $refObj.BASE.PARENT] $aggModel] {
            set refObj [c::getSuccessor $refObj]
        }

        # einfuegen
        change0::insert [c::get $refObj.BASE] $modelElem]
    }

}



proc edit::setValue {obj value} {
    foreach context [c::getContexts $obj] {
        if [procExists "edithook::setAttr_$context]"] {
            edithook::setAttr_$context $obj
        }
    }
    if {[equal [c::getValue $obj] $value]} {
        return
    }

    change0::setValue $obj $value
    if { [equal [c::getRole $obj] baseRef] } {
        return
    }
    if { [c::getModel $obj model] } {
        if { ![c::isReference $obj] || ![c::getModel $value modelValue] } {
            set modelValue $value
        }
        change0::setValue $model $modelValue
    }
}


proc edit::delete {elem} {
    
    set answer "no"
    
    foreach e $elem {
        if ![edit::canExtract $e] {
            if {[llength $elem] == 1} {
                CDlg::Error [i18n::text "Can't delete this object!"] "Delete"
            } else {
                CDlg::Error [i18n::text "Can't delete at least one selected object!"] "Delete"
            }
            return
        }
    }
    
    foreach e $elem {
        if [c::getModel $e model] {
            if {$answer == "no"} {
                set answer [ask [i18n::text "Delete model element as well?"]]
                if {$answer == "cancel"} {
                    return
                }
            }
        }
    }
    
    foreach e $elem {
        if [c::getModel $e model] {
        switch $answer {
            yes {                
                change0::delete $model
                change0::delete $e
            }
            no {
                change0::delete $e
            }
        }
        } else {
            change0::delete $e
        }
    }
}


# --- convenient edit functions ---------------------------------------------------------------


proc edit::setAttr {obj attr value} {
    set sub [c::getSubElement $obj $attr]
    setValue $sub $value
}

proc edit::setAttrAVO {attr value obj} {
    # fuer INSERT .. THEN ..
    setAttr $obj $attr $value
}


proc edit::insertConv {obj attr elem} {
    change0::insert [c::getSubElement $obj $attr] $elem
    return $elem
}


proc edit::setModel {objNode value} {
    change0::setValue [c::getSubElement $objNode baseRef] $value
}


proc edit::insertNew {aggNode class} {
    insert $aggNode [createObject $class [d::getFile $aggNode]]
}


proc edit::setClassOfChild {aggNode newClass} {
    if [equal [c::getClassOfChild $aggNode] $newClass] {
        return
    }
    set children [c::getList $aggNode.CHILDREN]
    if ![empty $children] {
        delete [lindex $children 0]
    }
    if ![empty $newClass] {
        insertNew [c::icBelow $aggNode] $newClass
    }
}


# --- check functions ------------------------------------------------------------------------


proc edit::canInsertType {insertionContext typeId} {
    if { [file::getReadonly [file::getFileOfNode $insertionContext]] } {
        return 0
    }
    return [c::canInsertType $insertionContext $typeId]
}


proc edit::getAcceptedType {insContext} {
    return [c::getAcceptedType [c::getParent $insContext]]
}


proc edit::canInsert {insertionContext newElem} {
    if { [file::getReadonly [file::getFileOfNode $insertionContext]] } {
        return 0
    }
    return [c::canInsert $insertionContext $newElem]
}


proc edit::canExtract {obj} {
    assert { [c::isLanguageObject $obj] }
    if { [file::getReadonly [file::getFileOfNode $obj]] } {
        return 0
    }
    return [expr { [c::canExtract $obj] && ![c::minMultiplicityIsOne [c::getParent $obj]] }]
}


# --- other functions ------------------------------------------------------------------------


proc edit::forceUpdate {obj} {
    file::forceUpdate [file::getFileOfNode $obj]
}


proc edit::getFile {node} {
    return [change::getFile $node]
}

