
namespace eval ::syncutil {}

# Dieses Modul stellt Hilfsfunktionen zur Verfgung, die zur Synchronisation
# des Strukturbaums benutzt werden knnen.


# --- Synchronisation von def/rep Listen -------------------------------------------------------


proc syncutil::createMissing0 {defList repList refAttr repCreationProc} {
    # Erzeugt fehlende Reprsentanten in $repList, die Objekte in $defList reprsentieren
    # $refAttr ist der Name des Attributs von Reprsentanten, das das reprsentierte Objekt referenziert.
    # Bekommt als Parameter ein Objekt aus $defList und muss ein entsprechendes Reprsentanten-Objekt
    # zurckliefern.

    # find missing
    foreach def $defList {
        set missing($def) 1
    }    
    foreach def [c::getList {$repList.CHILDREN.$refAttr.VALUE}] {
        set missing($def) 0
    }

    # createMissing
    foreach def $defList {
        if { $missing($def) } {
            set newobj [eval $repCreationProc $def]
            change0::insert [c::icBelow $repList] $newobj
            change0::setValue [c::get $newobj.$refAttr] $def
        }
    }
}


proc syncutil::createMissing {defList repList refAttr repCreationProc} {
    createMissing0 [c::getList {$defList.CHILDREN}] $repList $refAttr $repCreationProc
}


proc syncutil::deleteSuperfluous0 {defList repList refAttr} {
    # Lscht Objekte in $repList, die kein Objekt in $defList reprsentieren
    # $refAttr ist der Name des Attributs von Reprsentanten, das das reprsentierte Objekt referenziert.
   
    # find superfluous
    set superfluous ""
    foreach rep [c::getList {$repList.CHILDREN}] {
        set repRef [c::get $rep.$refAttr.VALUE]
        if { ![elementOf $repRef $defList] } {
            lappend superfluous $rep
        }
    }

    # delete superfluous
    foreach obj $superfluous {
        change0::delete $obj
    }
}


proc syncutil::deleteSuperfluous {defList repList refAttr} {
    deleteSuperfluous0 [c::getList {$defList.CHILDREN}] $repList $refAttr
}


proc syncutil::rightOrder {a b refAttr order} {
    # private
    return [expr [lsearch $order [c::get $a.$refAttr.VALUE]] <= [lsearch $order [c::get $b.$refAttr.VALUE]]]
}


proc syncutil::order0 {defList repList refAttr} {
    # ndert die Reihenfolge der Elemente in $repList, so dass sie der Reihenfolge der
    # reprsentierten Elemente in $defList entspricht.
    # $refAttr ist der Name des Attributs von Reprsentanten, das das reprsentierte Objekt referenziert.

    # find elements to re-insert
    set reInsert ""
    set order $defList
    set prev ""
    foreach rep [c::getList {$repList.CHILDREN}] {
        if { ![empty $prev] && ![rightOrder $prev $rep $refAttr $order] } {
            change0::extract $rep
            lappend reInsert $rep
        } else {
            set prev $rep
        }
    }

    # re-insert
    foreach obj $reInsert {
        foreach rep [c::getList {$repList.CHILDREN}] {
            if { ![rightOrder $rep $obj $refAttr $order] } {
                change0::insert $rep $obj
                break
            }
        }
    }
}


proc syncutil::order {defList repList refAttr} {
    order0 [c::getList {$defList.CHILDREN}] $repList $refAttr
}


proc syncutil::removeMultiple0 {repList refAttr} {
    # Lscht Elemente in $repList, die das gleiche Objekt reprsentieren wie ein anderes in $repList

    foreach obj $repList {
        set def [c::get {$obj.$refAttr.VALUE}]
        if { [info exists occured($def)] } {
            change0::delete $obj
        } else {
            set occured($def) 1
        }
    }
}


proc syncutil::removeMultiple {repList refAttr} {
    removeMultiple0 [c::getList {$repList.CHILDREN}] $refAttr
}


# --- Synchronisation base/derived -----------------------------------------------


proc syncutil::syncValue {obj} {
    # precondition: obj is a value object
    if { [c::getModel $obj model] } {
        change0::setValue $obj [c::get $model.VALUE]
    }
}


proc syncutil::syncReference {obj modelToRepName} {
    # precondition: obj is a reference object
    upvar $modelToRepName modelToRep    
    if { [c::getModel $obj model] } {
        if { ![equal [c::getOrNull $model.VALUE] [c::getOrNull $obj.VALUE.BASE]] } {
            if { ![info exists modelToRep([c::get $model.VALUE])] } {
                set newValue 0x0
            } else {
                set newValue [lindex $modelToRep([c::get $model.VALUE]) 0]
            }
            change0::setValue $obj $newValue
        }
    }
}


# xxx: sollte moveToCorrectList heissen
proc syncutil::moveToRightList {obj modelToRepName} {
    # precondition: obj is LanguageObject
    # postcondition: obj is in the "right" aggregate list, determined by the model
    upvar $modelToRepName modelToRep    
    if { [c::getModel $obj model] } {
        if { ![equal [c::get $obj.PARENT.PARENT.baseRef.VALUE] [c::get $model.PARENT.PARENT]] } {
            set destList [lindex $modelToRep([c::get $model.PARENT]) 0]
            change0::insert [c::icBelow $destList] $obj
        }
    }
}      


xn::proc syncutil::viewRepCreationProc {viewName model} {
    # returns the representation of object $model in views of type $viewName
    >> xn::assert { [c::isLanguageObject $model] }

    set className [c::getRepClass [c::getClassName $model] $viewName]
    return [change0::createObject $className [change::getFile $model]]
}


proc syncutil::createMissingRep {rep viewName} {
    # precondition: rep is an aggreate object
    # creates missing subelements of "rep".
    if { [c::getModel $rep model] } {
        createMissing $model $rep "baseRef" [list viewRepCreationProc $viewName]
    }
}


proc syncutil::orderRep {rep} {
    # precondition: rep is an aggreate object
    if { [c::getModel $rep model] } {
        order $model $rep baseRef
    }
}


proc syncutil::removeAbandoned {rep} {
    # precondition: rep is LanguageObject
    if { [c::getSubElement $rep baseRef -var baseRef] } {
        if { [nullptr [c::getValue $baseRef]] } {
            change0::delete $rep
            return 1
        }
    }
    return 0
}

# xxx: sollte zu SUB-Knoten gehoeren
proc syncutil::removeMultipleRep {rep modelToRepName} {
    # precondition: rep is LanguageObject
    upvar $modelToRepName modelToRep
    if { [c::getModel $rep model] } {
        if { [lsearch -exact $modelToRep($model) $rep]>0 } {
            change0::delete $rep
            return 1
        }
    }
    return 0
}
