
# xn - eXtended Notations
#
# Hier werden zusatznotationen definiert, mit denen 
# "schnerer" Code geschrieben werden kann.


namespace eval ::xn {

    set idCounter 0
    set debug 1

    proc setDebug {value} {
        assert { [member $value "0 1"] }
        set ::xn::debug $value
    }

    proc class {class args} {
        if { [xn::match "body" $args] } {
            Class subclass $class
        } elseif { [xn::match "inherits superclass body" $args] } {
            $superclass subclass $class
        } else {
            return -code error "Invalid parameters"
        }
        eval $body
    }
    
    proc method {name param body} {
        upvar 1 class class
        transformProc $param $body newParam newBody
        $class method $name $newParam "private \$object data\n$newBody"
    }

    proc classmethod {name param body} {
        upvar 1 class class
        $class classmethod $name $param $body
    }

    proc classvar {name value} {
        upvar 1 class class
        $class private $name $value
    }

    proc newId {} {
        variable idCounter
        return [incr idCounter]
    }

    proc new {class args} {
        set id $class[newId]
        eval [list $class new $id] $args
        return $id
    }

    proc code {args} {
        uplevel 1 [list namespace code $args]
    }

    proc call {cmd args} {
        return [eval $cmd $args]
    }

    proc assert {condition {message ""}} {
        set ok [uplevel 1 [list expr $condition]]
        if {!$ok} {
            if {$message eq ""} {
                set resolved [uplevel 1 [list subst -nocommands $condition]]
                return -code error "Assertion failed: $resolved"

            } else {
                set message [uplevel 1 [list subst $message]]
                return -code error "Assertion failed: $message"
            }
        }
    }

    proc substDebugTags {code} {
        if { $xn::debug } {
            set subSpec "\\1"
        } else {
            set subSpec "\\1\# "
        }
        set newCode [regsub -all -lineanchor {^( *)>> } $code $subSpec]
        return $newCode
    }


    proc optionsOk {list allowed messageVar} {
        upvar $messageVar message
        if { ([llength $list]%2) == 1 } {
            set message "Odd number of option-list elements"
            return 0
        }
        foreach {opt val} $list {
            if {[string index $opt 0] ne "-"} {
                set message "Option '$opt' doesn't start with a dash"
                return 0
            }
            if {[llength $allowed]>0 && ![member $opt $allowed]} {
                set message "Unknown option '$opt'"
                return 0
            }
        }
        return 1
    }


    proc transformProc {arglist body newArglistName newBodyName} {
        upvar 1 $newArglistName newArglist
        upvar 1 $newBodyName newBody

        set index [lsearch $arglist "-*"]
        if { $index == -1 } {
            set newArglist $arglist
            set newBody [substDebugTags $body]
        } else {
            set newArglist [lrange $arglist 0 [expr {$index-1}]]
            lappend newArglist "args"
            set nonPosArgs [lrange $arglist $index end]
            if { ![optionsOk $nonPosArgs "" msg] } {
                return -code error "Error in arglist: $msg"
            }
            set newBody ""
            if {$::xn::debug} {
                set allowedOptions ""
                foreach {opt val} $nonPosArgs {
                    lappend allowedOptions $opt
                }
                append newBody [list set __allowedOptions $allowedOptions] ";"
                append newBody {if {![::xn::optionsOk $args $__allowedOptions msg]} {error $msg}} ";"
            }
            set checkMandatory ""
            foreach {opt val} $nonPosArgs {
                if { $val eq "!" } {
                    append checkMandatory "if { !\[info exists [list [string range $opt 1 end]]\] } {error \"Missing mandatory non-positional arg '$opt'\"}" ";"
                }
                if { $val ne "!" && $val ne "?" } {
                    append newBody [list set [string range $opt 1 end] $val] ";"
                }
                set ::xn::optionMap($opt) [string range $opt 1 end]
            }
            append newBody {foreach {opt val} $args {set $::xn::optionMap($opt) $val}} ";"
            if {$::xn::debug} {
                append newBody $checkMandatory
            }
            append newBody "\n" [substDebugTags $body]
        }
    }

    proc proc {name arglist body} {
        transformProc $arglist $body newArglist newBody
        uplevel 1 [list ::proc $name $newArglist $newBody]
    }


    proc match {pattern value} {
        set n [llength $pattern]
        if { $n != [llength $value] } {
            return 0
        }
        set result ""
        for {set i 0} {$i<$n} {incr i} {
            set p [lindex $pattern $i]
            set v [lindex $value $i]            
            if {[string index $p 0] eq "!"} {
                if {[string range $p 1 end] ne $v} {return 0}
            } else {
                lappend result $p $v
            }
        }
        foreach {var val} $result {
            upvar 1 $var dest
            set dest $val
        }
        return 1
    }


    proc assign {pattern value} {
        set ok [uplevel 1 [list xn::match $pattern $value]]
        if {!$ok} {
            error "Pattern '$pattern' doesn't match '$value'"
        }
    }


    proc switch {value list} {
        foreach {pattern code} $list {
            if {[uplevel 1 [list xn::match $pattern $value]]} {
                set returnCode [catch [list uplevel 1 $code] result]
                return -code $returnCode -errorinfo $::errorInfo -errorcode $::errorCode $result
            }
        }
        error "Value doesn't match any pattern: '$value'"
    }


    proc throw {code {msg ""}} {
        if { $msg eq "" } {
            set msg "Internal errror (code: $code)"
        }
        error $msg "" $code
    }

    proc tryCatch {cmd list} {
        set returnCode [catch [list uplevel 1 $cmd] result]
        set savedErrorInfo $::errorInfo
        set savedErrorCode $::errorCode
        if {$returnCode == 1} {
            foreach {pattern code} $list {
                if {[string match $pattern $savedErrorCode]} {
                    upvar 1 errorMessage errorMessage
                    set errorMessage $result
                    set returnCode [catch [list uplevel 1 $code] result]
                    return -code $returnCode -errorinfo $::errorInfo -errorcode $::errorCode $result
                }
            }
        }
        return -code $returnCode -errorinfo $savedErrorInfo -errorcode $savedErrorCode $result
    }


    proc tryFinally {cmd finally} {
        set returnCode [catch [list uplevel 1 $cmd] result]
        set savedErrorInfo $::errorInfo
        set savedErrorCode $::errorCode
        set returnCode2 [catch [list uplevel 1 $finally] result2]
        if { $returnCode2 != 0 } {
            error "Error in finally block: $result2"
        }
        return -code $returnCode -errorinfo $savedErrorInfo -errorcode $savedErrorCode $result
    }
        

    proc member {elem set} {
        return [expr { [lsearch -exact $set $elem] != -1 }]
    }
}
